#!/usr/bin/python
# Messaging interface to G_message() and friends

import grass.lib.gis as libgis
from multiprocessing import Process, Lock, Pipe

def message_server(lock, conn):
    """!The GRASS messenger server run function
    
       @param lock A multiprocessing lock
       @param conn A multiprocessing pipe 
       
       This function will use the G_* message C-functions from grass.lib.gis
       to provide an interface to the GRASS C-messaging system.
       
       The data that is recived must provide an identifierstring to specify 
       which C-function should be called. The following identifers are 
       supported:

       - "INFO"       Prints an info message, see G_message() for details
       - "IMPORTANT"  Prints an important info message, 
                      see G_importnant_message() for details
       - "VERBOSE"    Prints a verbose message if the verbosity level is 
                      set accordingly, see G_verbose_message() for details
       - "WARNING"    Prints a warning message, see G_warning() for details
       - "ERROR"      Prints an error message witha leading "ERROR: " string, 
                      see G_important_message() for details
       - "PERCENT"    Prints a percent value based on three values: n, d and s
       - "STOP"       Stops the server function and closes the pipe
       - "FATAL"      Calls G_fatal_error(), this functions is only for 
                      testing purpose
        
       Data to be recived via pipe should be a list of values:
       - Messages: ["INFO|VERBOSE|WARNING|ERROR", "MESSAGE"]
       - Debug: ["DEBUG", level, "MESSAGE"]
       - Percent: ["PERCENT", n, d, s]
       - Fatal: ["FATAL",]
    """
    while True:
        # Avoid busy waiting
        conn.poll(None)
        data = conn.recv()
        message_type = data[0]
        
        # Only one process is allowed to write to stdout
        lock.acquire()
        # Stop the pipe and the infinite loop
        if message_type == "STOP":
            conn.close()
            lock.release()
            return
        # This is for testing only
        elif data[0] == "FATAL":
            libgis.G_fatal_error("this is a fatal error")        
                    
        message = data[1]
        
        if message_type == "PERCENT":
            n = int(data[1])
            d = int(data[2])
            s = int(data[3])
            libgis.G_percent(n, d, s)
        elif message_type == "INFO":
            libgis.G_message(message) 
        elif message_type == "VERBOSE":
            libgis.G_verbose_message(message)  
        elif message_type == "DEBUG":
            level = data[1]
            message = data[2]
            libgis.G_debug(level, message)  
        elif message_type == "IMPORTANT":
            libgis.G_important_message(message)  
        elif message_type == "WARNING":
            libgis.G_warning(message)
        elif message_type == "ERROR":
            libgis.G_important_message("ERROR: %s"%message)

        lock.release()
            
class Messenger(object):
    """!Message interface to the GRASS message C-functions
        
       Usage:
    
       @code
       msgr = Messenger()
       msgr.message("message")
       msgr.verbose("verbose message")
       msgr.important("important message")
       msgr.percent(1, 1, 1)
       msgr.debug(0, "debug 0")
       msgr.warning("Ohh")
       msgr.error("Ohh no")
       msgr.stop()
       @endcode
    """
    def __init__(self):
        self.start_server()
            
    def start_server(self):
        self.client_conn, self.server_conn = Pipe()
        self.lock = Lock()
        self.server = Process(target=message_server, args=(self.lock, 
                                                          self.server_conn))
        self.server.daemon = True
        self.server.start()
        
    def _check_restart_server(self):
        """!Restart the server if it was terminated
        """
        if self.server.is_alive() is True:
            return
        self.client_conn.close()
        self.server_conn.close()
        self.start_server()
        self.warning("Needed to restart the messenger server")
        
    def message(self, message):
        """!Send a message to stdout
        
           G_message() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["INFO", message])
        
    def verbose(self, message):
        """!Send a verbose message to stdout
        
           G_verbose_message() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["VERBOSE", message])
        
    def important(self, message):
        """!Send an important message to stdout
        
           G_important_message() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["IMPORTANT", message])
        
    def warning(self, message):
        """!Send a warning message to stdout
        
           G_warning() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["WARNING", message])
        
    def error(self, message):
        """!Send an error message to stdout
        
           G_important_message() with an additional "ERROR:" string at 
           the start will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["ERROR", message])
        
    def debug(self, level, message):
        """!Send a debug message to stdout
        
           G_debug() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["DEBUG", level, message])
    
    def percent(self, n, d, s):
        """!Send a percentage to stdout
        
           G_percent() will be called in the messenger server process
        """
        self._check_restart_server()
        self.client_conn.send(["PERCENT", n, d, s])
            
    def stop(self):
        """!Stop the messenger server and close the pipe
        """
        self.client_conn.send(["STOP",])
        self.server.join(1)
        if self.server.is_alive():
            self.server.terminate()
        self.client_conn.close()
    
    def test_fatal_error(self):
        """!Force the messenger server to call G_fatal_error()
        """
        import time
        self._check_restart_server()
        self.client_conn.send(["FATAL",])
        time.sleep(1)
     
if __name__ == "__main__":
    msgr = Messenger()
    msgr.message("message")
    msgr.verbose("verbose message")
    msgr.important("important message")
    msgr.test_fatal_error()
    msgr.percent(1, 1, 1)
    msgr.debug(0, "debug 0")
    msgr.warning("Ohh")
    msgr.debug(1, "debug 1")
    msgr.error("Ohh no")
    msgr.stop()
    # This should result in:
    """
    message
    important message
    ERROR: this is a fatal error
    WARNING: Needed to restart the messenger server
     100%
    D0/0: debug 0
    WARNING: Ohh
    ERROR: Ohh no
    """
    # Test of the percentage creation
    msgr = Messenger()
    num = 100000
    for i in range(num):
        msgr.percent(i, num, 10)
    msgr.stop()
    